/*----------------------------------------------------------------------
 *
 *      Title:          CDL main module
 *
 *      Description:
 *              This module provides all of the required functions to drive
 *      the command definition language (CDL).
 *
 *      Author:   Richard J. Harrison
 *
 *      Date  :   30 November 1989
 *
 *      Rev:    1.3
 *      1.1 R.J.Harrison        Release version                 30 Nov 89
 *      1.2 R.J.Harrison        Add quotes & escaped chars.      2 Mar 90
 *      1.3 R.J.Harrison        Fix crud left in string bug      7 Mar 90
 *----------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>

#define CDL_MODULE TRUE

#include "zpl.h"

static TY_int finished_flag;

static TOK *tok, *root_tok;
static TY_int num_of_commands, max_commands, interactive_mode = TRUE;

void init_cdl (TOK * tok_storage, TY_int mc)
{
    root_tok = tok = tok_storage;
    max_commands = mc;
    num_of_commands = 0;
}

void set_cdl (TOK_ROOT * tr)
{
    tok = tr->table;
    max_commands = tr->max_commands;
    num_of_commands = tr->num_commands;
}

void store_cdl (TOK_ROOT * tr)
{
    tr->table = root_tok;
    tr->max_commands = max_commands;
    tr->num_commands = num_of_commands;
}


static TY_int stoi (register char **instr)
{
    register TY_int num = 0;
    register char *str;
    TY_int sign = 1;

    str = *instr;

if(!str)
return 0;

    while (*str && 
          (*str == ' ' || *str == '\t' || *str == '\n') )
	str++;

    if(!*str)
        return 0;

    if (*str == '-')
    {
	sign = -1;
	str++;
    }
    if (*str == '0')
    {
	++str;
	if (*str == 'x' || *str == 'X')
	{
	    str++;
	    while (('0' <= *str && *str <= '9') ||
		   ('a' <= *str && *str <= 'f') ||
		   ('A' <= *str && *str <= 'F'))
	    {
		num *= 16;
		num += ('0' <= *str && *str <= '9') ?
		    *str - '0' :
		    toupper (*str) - 'A' + 10;
		str++;
	    }
	}
	else
	{
	    while ('0' <= *str && *str <= '7')
	    {
		num *= 8;
		num += *str++ - '0';
	    }
	}
    }
    else
    {
	while ('0' <= *str && *str <= '9')
	{
	    num *= 10;
	    num += *str++ - '0';
	}
    }
    *instr = str;
    return (num * sign);
}


/*
 * this function skips over any text contained within quotes
 */
static TY_int is_quote (char ch)
{
    return (ch == '"' || ch == '\'');
}

char *skip_quotes (ch)
char *ch;
{
    char quote;

    if (is_quote (*ch))
    {
	quote = *ch++;
	while (*ch && *ch != quote)
	    ch++;
	if (*ch)
	    ch++;
    }
    return (ch);
}

char *check_quotes (pl)
char *pl;
{
    if (is_quote (*pl))
    {
	pl = skip_quotes (pl);
    }
    return (pl);
}


static TY_int int_value_required (PARAM * paramp)
{
    if (paramp->type == INTEGER)
	return (TRUE);
    else
	return (FALSE);
}

static TY_int value_required (PARAM * paramp)
{
    switch (paramp->type)
    {
    case INTEGER:
    case CHARACTER:
    case STRING:
	return (TRUE);
    }
    return (FALSE);
}


/*
 * When a parameter has been found, this function will take the
 * appropriate action
 */

static char *setarg (PARAM * paramp, char *linep)
{
    switch (paramp->type)
    {
    case INTEGER:
	*paramp->variable = stoi (&linep);
	break;

    case FBOOLEAN:
	*paramp->variable = 1;
	break;

    case TBOOLEAN:
	*paramp->variable = 0;
	break;

    case SBOOLEAN:
	*paramp->variable = ~*paramp->variable;
	break;

    case CHARACTER:
	*paramp->variable = *linep++;
	break;

    case STRING:
	{

	    /*
	     * We have to allocate some storage to save this string as
	     * the variable we are setting is only a pointer. It is
	     * possibly bad practice to malloc something which we have no
	     * intention of releasing but I don't care about such social
	     * niceities
	     */
	    if (linep)
	    {
		char *save = (char *) malloc (strlen (linep) + 2);
		TY_int len = strlen (linep);
		if (save)
		{
		    if (is_quote (*linep))
		    {
			if (is_quote (linep[len - 1]))
			    len--;
			linep++;
			len--;
		    }
		    strncpy (save, linep, len);
		    save[len] = 0;
		    *(char **) paramp->variable = save;
		}
		else
		{
		    error_log ("CDL: Unable to allocate storage for command item %s", linep);
		}
		break;
	    }
	}
    default:
	error_log ("CDL: Bad CDL type %ld", paramp->type);
	break;
    }
    return (linep);
}

static char *strip_spaces (char *ptr)
{
    register char *ptr1 = ptr, quote = 0;

    while (*ptr)
    {
	if (is_quote (*ptr))
	{
	    if (quote && *ptr == quote)
	    {
		quote = 0;
	    }
	    else
	    {
		quote = *ptr;
	    }
	}
	if (*ptr != ' ' || quote)
	    *ptr1++ = *ptr;
	ptr++;
    }
    *ptr1 = 0;
    return (ptr);
}

TOK *add_command (command, func)
char *command;
void (*func) ();
{
    TOK *ptr;
    ptr = &tok[num_of_commands++];
    strncpy (ptr->command, command, COMLEN - 1);
    strncpy (ptr->min_abb, command, COMLEN - 1);
    ptr->func = func;
    return (ptr);
}

void add_parameter (tok, parameter, variable, type, flags)
TOK *tok;
PARAM_TYPE type;
char *parameter;
void *variable;
TY_int flags;
{
    PARAM *param;

    param = (PARAM *) malloc (sizeof (PARAM));

    if (!param)
    {
	error_log ("CDL: Cannot allocate %ld bytes for parameter", sizeof (PARAM));
    }
    else
    {
	if (tok->number_of_parameters >= MAXPARAM - 1)
	{
	    error_log ("CDL Internal error, too many parameters for %s", tok->command);
	    return;
	}
	else
	{
	    tok->parameters[tok->number_of_parameters] = param;
	    strncpy (param->parameter, parameter, PARAMLEN);
	    strncpy (param->min_abb_p, parameter, COMLEN);
	    param->type = type;
	    param->flags = flags;
	    param->variable = variable;
	    tok->number_of_parameters++;
	}
    }
}

static find_first_non_match (char *s1, char *s2)
{
    TY_int i = 0;
    while (*s1 && *s2 && *s1++ == *s2++)
	i++;
    return (i);
}

static void abbreviate_parameters (TOK * atok)
{
    TY_int ap, cp, min_length, first_non_equal;
    char *s1, *s2;
    TOK *ctok;

    for (ap = 0; ap < atok->number_of_parameters; ap++)
    {
	min_length = 0;
	s1 = atok->parameters[ap]->parameter;

#if ABBREVIATE_ALL_PARAMETERS
	for (c = 0; c < num_of_commands; c++)
#endif
	{
#if ABBREVIATE_ALL_PARAMETERS
	    ctok = &tok[c];
#else
	    ctok = atok;
#endif

	    for (cp = 0; cp < ctok->number_of_parameters; cp++)
	    {
		s2 = ctok->parameters[cp]->parameter;

		if (s1 != s2)
		{
		    first_non_equal = find_first_non_match (s1, s2);
		    if (first_non_equal > min_length)
			min_length = first_non_equal;
		}
	    }
	}
	atok->parameters[ap]->min_abb_p[min_length + 1] = 0;
    }
}

void abbreviate_commands ()
{
    TY_int c, i, min_length, first_non_equal;

    for (i = 0; i < num_of_commands; i++)
    {
	min_length = 0;
	for (c = 0; c < num_of_commands; c++)
	{
	    if (c != i)
	    {
		first_non_equal = find_first_non_match (tok[i].command,
							tok[c].command);
		if (first_non_equal > min_length)
		    min_length = first_non_equal;
	    }
	}
	tok[i].min_abb[min_length + 1] = 0;
	abbreviate_parameters (&tok[i]);
    }
}

static command_match (char *command, int maxlen,
		      char *min_abb, char *full)
{
    change_case_lower (command);
    if (!strncmp (min_abb, command, strlen (min_abb)))
    {

	/*
	 * The minimum abbreviation is good, but ensure that any extra
	 * characters given after this are also accurate
	 */
	if (!strncmp (full, command, maxlen))
	{
	    return (TRUE);
	}
    }
    return (FALSE);
}

/*
 * This function is rather like the unos 'strpbrk' function, except
 * that it ignores text within quotes, and it allows seperation
 * characters to be 'escaped'
 */

char *cdl_strpbrk (pl, break_list)
char *pl, *break_list;
{
    static char *ptr;
    if (pl)
    {
	while (*pl)
	{
	    ptr = break_list;
	    pl = check_quotes (pl);
	    while (*ptr)
	    {
		if (*pl == *ptr)
		{
		    if (pl[1] != *ptr)
		    {
			return (pl);
		    }
		    else
		    {
			ptr = NULL;
			pl += 2;
		    }
		}
		else
		{
		    ptr++;
		}
	    }
	    pl++;
	}
    }
    return (NULL);
}


char *cdl_strtok (param, break_list)
char *param, *break_list;
{
    static char *pl, *zz, *zzt;
    if (param)
    {
	pl = param;
    }
    zz = cdl_strpbrk (pl, break_list);
    if (pl)
    {
	zzt = pl;
	if (zz)
	{
	    *zz++ = 0;
	}
	pl = zz;
	return (zzt);
    }
    else
    {
	zz = pl;
	pl = 0;
	return (zz);
    }
}

static char *param_prompt (char *param)
{
    static char prompt[200];
    sprintf (prompt, "%s", param);
    prompt[0] = toupper (prompt[0]);
    return (prompt);
}

static char *check_parameters (TOK * tok, char *params, long param)
{
    TY_int i = 0, pfound = 0, maxlen;
    static char *value, *par, static_params[200], error_line[80], *hack_next;
static char new_value[200];
    TY_int positional_index = 0;

    /*
     * Clear all the parameters 'present' flag. This is to allow us to check
     * the required parameters at the end... Any parameters which are
     * required and not present will be prompted for :-)
     */

    tok->flags &= ~PARAM_PRESENT;

    for (i = 0; i < tok->number_of_parameters; i++)
	tok->parameters[i]->flags &= ~PARAM_PRESENT;

    if (params && *params)
    {

	memset (static_params, 0, sizeof (static_params));
	strcpy (static_params, params);

	/*
	 * strip_spaces(static_params); 
	 */

	par = cdl_strtok (static_params, ",/ ");


	while (par)
	{
	    hack_next = 0;

	    if (value = strchr (par, '='))
	    {
		*value++ = 0;
	    }
	    i = pfound = 0;

	    maxlen = strlen (par);

	    while (i < tok->number_of_parameters && !pfound)
	    {

		if (!(tok->parameters[i]->flags & PARAM_POSITIONAL) &&
		    command_match (par, maxlen,
				   tok->parameters[i]->min_abb_p,
				   tok->parameters[i]->parameter))
		{

		    /*
		     * we have found a parameter which is understood so
		     * process it
		     */
		    if (!value && value_required (tok->parameters[i]))
		    {
			value = cdl_strtok (NULL, ",/ ");
			if (value && !isdigit (*value) && int_value_required (tok->parameters[i]))
			    hack_next = value;
			else
			    hack_next = NULL;
		    }
		    if ((tok->parameters[i]->flags & PARAM_VALUE_REQUIRED) && (!value || !*value))
		    {
if(interactive_mode)
{
			*new_value = 0;
			strinput (0, new_value, param_prompt (tok->parameters[i]->parameter), sizeof (new_value));
			setarg (tok->parameters[i], new_value);
}
else
	    {
		sprintf (error_line, "Parameter '%s' has no value", tok->parameters[i]->parameter);
		return (error_line);
	    }
		    }
		    else
			setarg (tok->parameters[i], value);
		    tok->parameters[i]->flags |= PARAM_PRESENT;
                    tok->flags |= PARAM_PRESENT;

		    pfound = TRUE;
		}
		i++;
	    }

	    if (!pfound && maxlen && positional_index < tok->number_of_parameters &&
		tok->parameters[positional_index]->flags & PARAM_POSITIONAL)
	    {
		i = positional_index++;
		pfound = TRUE;
		tok->parameters[i]->flags |= PARAM_PRESENT;
                tok->flags |= PARAM_PRESENT;

		/*
		 * we have found a parameter which is understood so process
		 * it
		 */
		setarg (tok->parameters[i], par);
	    }

	    if (!pfound && par && *par)
	    {
		sprintf (error_line, "Unknown parameter '%s'", par);
		return (error_line);
	    }
	    if (hack_next)
		par = hack_next;
	    else
		par = cdl_strtok (0, " ,/");
	}
    }
    for (i = 0; i < tok->number_of_parameters; i++)
    {
	if (tok->parameters[i]->flags & PARAM_REQUIRED &&
	    !(tok->parameters[i]->flags & PARAM_PRESENT))
	{
	    *new_value = 0;
if(interactive_mode)
{
	    if (strinput (0, new_value, param_prompt (tok->parameters[i]->parameter), sizeof (new_value)))
		setarg (tok->parameters[i], new_value);
	    if (!*new_value)
	    {
		sprintf (error_line, "As value not given for parameter '%s', command aborted",
			 tok->parameters[i]->parameter);
		return (error_line);
	    }
}
else
	    {
		sprintf (error_line, "Required parameter '%s' missing", tok->parameters[i]->parameter);
		return (error_line);
	    }
	    tok->parameters[i]->flags |= PARAM_PRESENT;
            tok->flags |= PARAM_PRESENT;
	}
    }

    if( (tok->flags & PARAM_REQUIRED) && !(tok->flags & PARAM_PRESENT))
        return "At least one parameter must be specified";

    return (NULL);
}

char *check_command (char *input, TY_int param)
{
    TY_int i = 0, found = FALSE;
    char *command, *params, *ch;

    if (params = strpbrk (input, " /,"))
    {
	*params++ = 0;
    }
    else
    {
	params = NULL;
    }
    command = input;

    if (command == NULL)
	return (NULL);

    while (i < num_of_commands && !found)
    {
	if (command_match (command, strlen (command), tok[i].min_abb, tok[i].command))
	{

	    /*
	     * Now see if we have any parameters. if a parameter has been
	     * defined which requires the setting of a variable, and there is
	     * no function associated with this command, don't call it
	     */
	    ch = check_parameters (&tok[i], params, param);
	    if (ch)
		return (ch);
	    if (tok[i].func)
	    {
		(*tok[i].func) (&tok[i], param);
	    }
	    found = TRUE;
	}
	i++;
    }
    return (found ? NULL : "Unknown command");
}

TY_int find_command (TOK * tok, char *input, TY_int num_of_commands)
{
    TY_int i = 0, found = FALSE;
    char *command, *params;

    if (params = strpbrk (input, " /,"))
    {
	*params++ = 0;
    }
    else
    {
	params = NULL;
    }
    command = input;

    if (command == NULL)
	return (NULL);

    while (i < num_of_commands && !found)
    {
	if (command_match (command, strlen (command), tok[i].min_abb, tok[i].command))
	    return (i);
	i++;
    }
    return (0);
}

static char *op_var (PARAM * param)
{
    static char op[30];

    memset (op, 0, sizeof (op));

    switch (param->type)
    {
    case INTEGER:
	sprintf (op, "INT=%5ld", *(param->variable));
	break;

    case TBOOLEAN:
	sprintf (op, "TB =%5s", *(param->variable)
		 ? "True" : "False");
	break;

    case FBOOLEAN:
	sprintf (op, "FB =%5s", *(param->variable)
		 ? "True" : "False");
	break;

    case CHARACTER:
	sprintf (op, "CH =%5c", *(param->variable));
	break;

    case STRING:
	sprintf (op, "ST =<%s>", *(char **) param->variable);
	break;
    }
    return (op);
}

char *CDL_describe_var (param)
PARAM *param;
{
    switch (param->type)
    {
    case INTEGER:
	return ("Numeric");

    case TBOOLEAN:
	return ("Off-Switch");

    case FBOOLEAN:
	return ("Switch");

    case CHARACTER:
	return ("Character");

    case STRING:
	return ("String");
    }
    return ("Unknown");
}

void dump_cdl_table (void (*opl) (TY_int, char *), TY_int par)
{
    TY_int c, cp;
    TOK *ctok;
    char line[120];

    for (c = 0; c < num_of_commands; c++)
    {
	ctok = &tok[c];
	sprintf (line, "\nCommand: %-20s, %-20s %3ld parameters\n",
		 ctok->command, ctok->min_abb, ctok->number_of_parameters);
	(*opl) (par, line);

	for (cp = 0; cp < ctok->number_of_parameters; cp++)
	{
	    PARAM *p = ctok->parameters[cp];
	    sprintf (line, "\tP%3ld: %-20s %-20s[%s]\n", cp, p->parameter, p->min_abb_p,
		     op_var (p));
	    (*opl) (par, line);
	}
    }
}

void definition_end ()
{
    finished_flag = TRUE;
}


void set_cdl_interactive(TY_int interactive)
{
interactive_mode = interactive;
}

#ifdef SCRIPT
void process_cdl_file (filename)

char *filename;
{
    static char command[256];
    FILE *conf;
    TY_int errors_occured = 0;

    if ((conf = fopen (filename, "r")))
    {
	while ((fgetline (conf, command, sizeof (command)) != EOF) &&
	       !finished_flag)
	{
	    if (command[0] != '#' && command[0] != '\n')
	    {
		if (strlen (command) && check_command (command))
		{
		    errors_occured++;
		    fprintf (stderr, "nicola:Unknown configuration command \\%s\\\n", command);
		}
	    }
	    memset (command, 0, sizeof (command));
	}
	fclose (conf);
	return (errors_occured);	/*
					 * tell user of errors 
					 */
    }
    return (-1);		/*
				 * tell user file not found 
				 */
}

#endif

