/*---------------------------------------------------------------------------
 *
 *      Title:          NICOLA Database - GranAlloc routines
 *
 *      File:           qdb_gi.c
 *
 *      Date:           1991
 *
 *      Description:
 *              This module maintains the 'granalloc' lists used withing
 *      the system. A 'granalloc' list is a chunk of dynamic memory which
 *      has a defined size, a granularity (size of increment) and a maximum
 *      current size. Whenever an item is added which would cause the maximum
 *      size to be exceeded then a new chunk of size 'granularity' is added
 *      onto the array
 *
 *      Routines are present for reading, writing, initialising, disposing,
 *      maintaining and deleting.
 *
 *
 *      Rev: 1.0
 *      1.0     R.J.Harrison    Implement                            1991
 *
 *      Copyright  1993 R.J.Harrison
 *---------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "zpl.h"

void init_gi_item (QDB_GRAN_ITEM * mf, TY_int elem_size, TY_int gran)
{
    mf->size = mf->max_size = 0;
    mf->elem_size = elem_size;
    mf->gran = gran;
    mf->init = TRUE;
}

void *gran_alloc (TY_int gran, TY_int cur_size, TY_int elem_size)
{
    void *c = calloc ((cur_size + gran), elem_size);
    if (!c)
	error_log ("Couldn't get %ld bytes", (cur_size + gran) * elem_size)
	    ;
    else
	memset (c, 0, (cur_size + gran) * elem_size);
    return (c);
}

/*
 * First check to see if we've reached the end of the
 * current list, if we have then allocate enough space
 * for a new list, copy the old, adjust the size and
 * free the old list.
 * This routine will return a pointer to a valid block, whether or not
 * the valid block is the original or the new block...
 */

void *check_gran_alloc (void *cp, TY_int size, TY_int * max_size, TY_int gran, TY_int elem_size)
{
    if (size >= *max_size)
    {
	void *new_index = gran_alloc (gran, size, elem_size);
	if (new_index)
	{
	    if (cp)
	    {
		memcpy (new_index, cp, size * elem_size);
		MY_ov_check_memory ();
		free (cp);
	    }
	    *max_size += gran;
	    return (new_index);
	}
    }
    return (cp);
}

void check_gi_alloc (QDB_GRAN_ITEM * gi)
{
    gi->ptr = check_gran_alloc (gi->ptr, gi->size, &gi->max_size, gi->gran, gi->elem_size);
}


void QDB_read_gi (QDB_GRAN_ITEM * qit, char *filename, TY_int magic_no)
{
    FILE *f;

    if (filename)
    {
	if (f = fopen (filename, "r"))
	{
	    check_file_id (f, filename, magic_no);

	    fread ((char *) &qit->size, 1, sizeof (qit->size), f);
	    if (qit->size)
	    {
		check_gi_alloc (qit);

		if (qit->ptr)
		{
		    fread (qit->ptr, qit->size, qit->elem_size, f);
		}
		else
		{
		    error_log ("Unable to get memory for GI_item (file %s, size %ld)", filename, qit->size);
		}
	    }
	    fclose (f);
	}
    }
}

void QDB_update_gi (QDB_GRAN_ITEM * qit, char *filename, TY_int magic_no)
{
    FILE *f;

    if (qit && qit->changed)
    {
	if (filename)
	{
	    if (!qit->ptr || !qit->size)
	        unlink (filename);
	    else if (f = fopen (filename, "w"))
	    {
		put_file_id (f, magic_no);

		if (qit->init)
		{
		    fwrite ((char *) &qit->size, 1, sizeof (qit->size), f);

		    if (qit->ptr)
		    {
			fwrite ((char *) qit->ptr, qit->size, qit->elem_size, f);
		    }
		}
		fclose (f);
	    }
	}
    }
}

/*
 * Delete items from the list start.
 * The list will not be compacted.
 */
void QDB_delete_gi (QDB_GRAN_ITEM * qit, TY_int start, TY_int number)
{
    if (qit->init && qit->ptr)
    {
	qit->changed = TRUE;
	if (start >= qit->size)
	{
	    qit->size = start - 1;
	    return;
	}
	if (qit->size < number)
	{
	    qit->size = 0;
	    memset (qit->ptr, 0, qit->size * qit->elem_size);
	}
	else
	{
	    char *ptr = (char *) qit->ptr;
	    char *st = ptr + (start * qit->elem_size);
	    char *sc = ptr + ((start + number) * qit->elem_size);
	    TY_int number_to_copy = qit->size - (start + number);
	    memcpy (st, sc, number_to_copy * qit->elem_size);
	    qit->size -= number;
#ifdef CLEAN
	    sc = ptr + (qit->size * qit->elem_size);
	    memset (sc, 11, (qit->max_size - qit->size) * qit->elem_size);
#endif
	}
    }
}

void free_gi_item (QDB_GRAN_ITEM * qit)
{
    if (qit->init)
    {
	qit->size = qit->max_size = 0;
	if (qit->ptr)
	    free (qit->ptr);
	qit->ptr = 0;
	qit->init = 0;
	qit->changed = TRUE;
    }
}

TY_int QDB_find_gi_index (QDB_GRAN_ITEM * qit, void *item)
{
    return ((TY_int) item - (TY_int) qit->ptr) / (TY_int) qit->elem_size;
}

void QDB_delete_gi_item (QDB_GRAN_ITEM * qit, void *item)
{
    TY_int ind = QDB_find_gi_index (qit, item);
    QDB_delete_gi (qit, ind, 1);
}

void QDB_mark_gi_changed (QDB_GRAN_ITEM * qit)
{
    qit->changed = TRUE;
}
