#ifdef M_AMIGA
/* 
   appkeymap.c - subroutines to copy the default keymap, modify the copy

   Usage: struct KeyMap *appkeymap = CreateAppKeyMap();
   void DeleteAppKeyMap(struct KeyMap *appkeymap);

   This example modifies the copied keymap by unmapping all of the
   numeric keypad keys.  This creates a good keymap for use with either
   keymap.library MapANSI() or commodities InvertString().
   If you used a default keymap with the above functions, numeric keypad
   raw key values would be returned for keys which are available with
   fewer keypresses on numeric pad than on the normal keyboard.
   It is generally preferable to have the normal keyboard raw values
   since many applications attach special meanings to numeric pad keys.
   The AlterAppKeyMap() routine in this module could easilty be modified
   to instead set new values for numeric pad or function keys.

   IMPORTANT: Do Not Use SetKeyMapDefault() unless you are a system
   preferences editor OR an application that takes over the machine.
   If you want to use a customized keymap such as this in your
   application, open your own Intuition window, attach a console
   device to it, and then use the console device SETKEYMAP command
   to set your console device unit to your custom keymap.

 */

#include <stdlib.h>

#include <exec/types.h>
#include <exec/memory.h>
#include <devices/console.h>
#include <devices/keymap.h>

#include <clib/exec_protos.h>
#include <clib/keymap_protos.h>
#include <clib/alib_protos.h>

/* Count of elements in keymap arrays */
#define MAP_SIZE  64
#define TYPE_SIZE 64
#define CAPS_SIZE 8
#define REPS_SIZE 8

/* We allocate our Lo and Hi array pairs each as a single array */
struct KeyMapArrays
{
    ULONG          LHKeyMap[MAP_SIZE << 1];
    UBYTE          LHKeyMapTypes[TYPE_SIZE << 1];
    UBYTE          LHCapsable[CAPS_SIZE << 1];
    UBYTE          LHRepeatable[REPS_SIZE << 1];
};

static struct KeyMapArrays *karrays;
static struct KeyMap *defkeymap, *appkeymap;
static LONG    mapsize;

/* local functions */
static void 
CopyKeyMap (struct KeyMap *s, struct KeyMap *d)
{
    UBYTE         *bb;
    ULONG         *ll;
    int            i;

/* Copy keymap s (source) to keymap d (dest) */
    for (i = 0, ll = s->km_LoKeyMap; i < MAP_SIZE; i++)
	d->km_LoKeyMap[i] = *ll++;
    for (i = 0, ll = s->km_HiKeyMap; i < MAP_SIZE; i++)
	d->km_HiKeyMap[i] = *ll++;

    for (i = 0, bb = s->km_LoKeyMapTypes; i < TYPE_SIZE; i++)
	d->km_LoKeyMapTypes[i] = *bb++;
    for (i = 0, bb = s->km_HiKeyMapTypes; i < TYPE_SIZE; i++)
	d->km_HiKeyMapTypes[i] = *bb++;

    for (i = 0, bb = s->km_LoCapsable; i < CAPS_SIZE; i++)
	d->km_LoCapsable[i] = *bb++;
    for (i = 0, bb = s->km_HiCapsable; i < CAPS_SIZE; i++)
	d->km_HiCapsable[i] = *bb++;

    for (i = 0, bb = s->km_LoRepeatable; i < REPS_SIZE; i++)
	d->km_LoRepeatable[i] = *bb++;
    for (i = 0, bb = s->km_HiRepeatable; i < REPS_SIZE; i++)
	d->km_HiRepeatable[i] = *bb++;
}

struct Library *KeymapBase = NULL;

struct KeyMap *
CreateAppKeyMap ( void (*AlterAppKeyMap) (struct MY_KeyMap *) )
{
    if (KeymapBase = OpenLibrary ("keymap.library", 37L))
    {
	/* Get a pointer to the keymap which is set as the system default */
	if (defkeymap = AskKeyMapDefault ())
	{
	    /* Allocate our KeyMap structures and arrays */
	    mapsize = sizeof (struct KeyMap) + sizeof (struct KeyMapArrays);
	    if (appkeymap = calloc (mapsize, 1))
	    {
		/* Init our appkeymap fields to point to our allocated arrays.
		 * Here the  + 1  means  + sizeof(struct KeyMap)  because
		 * we are adding 1 to a pointer to a struct KeyMap.
		 * Each LH array contains a Lo and a Hi array.
		 */
		karrays = (struct KeyMapArrays *) (appkeymap + 1);
		appkeymap->km_LoKeyMap = &karrays->LHKeyMap[0];
		appkeymap->km_HiKeyMap = &karrays->LHKeyMap[MAP_SIZE];
		appkeymap->km_LoKeyMapTypes = &karrays->LHKeyMapTypes[0];
		appkeymap->km_HiKeyMapTypes = &karrays->LHKeyMapTypes[TYPE_SIZE];
		appkeymap->km_LoCapsable = &karrays->LHCapsable[0];
		appkeymap->km_HiCapsable = &karrays->LHCapsable[CAPS_SIZE];
		appkeymap->km_LoRepeatable = &karrays->LHRepeatable[0];
		appkeymap->km_HiRepeatable = &karrays->LHRepeatable[REPS_SIZE];

		/* Copy the user's default system keymap arrays to our appkeymap
		 * arrays to get the proper keymappings for the user's keyboard.
		 */
		CopyKeyMap (defkeymap, appkeymap);

		/* Now make our changes to our appkeymap */
		AlterAppKeyMap ((struct MY_KeyMap *)appkeymap);
	    }
	}
	CloseLibrary (KeymapBase);
    }
    return (appkeymap);
}

#endif
