/*---------------------------------------------------------------------------
 *
 *      Title:          NICOLA Preferences Database
 *
 *      File:           prefs_db.c
 *
 *      Date:           August 1992
 *
 *      Description:
 *              This module is responsible for storing a preferences data structure,
 *      which consists of smaller abstracted data entities, the only
 *      description being the amount of data to save
 *
 *      Rev: 1.0
 *      1.0     R.J.Harrison    Implement                            August 1992
 *
 *      Copyright  1993 R.J.Harrison
 *---------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/types.h>

#include "zpl.h"

static void add_preference (PREFERENCES * prefs, PREFERENCE_ENTRY * pe)
{
    PREFERENCE_LIST *rr, *new_rr;

    rr = prefs->pref_list;
    while (rr && rr->next)
    {
	rr = rr->next;
    }

    if ((new_rr = calloc (1, sizeof (*new_rr))))
    {
	if (rr)
	    rr->next = new_rr;
	else
	    prefs->pref_list = new_rr;
	new_rr->data = pe;
    }
}


TY_int write_pref (PREFERENCE_ENTRY * pe, FILE * f)
{
    if (!pe)
	return (FALSE);
    write_text (f, pe->name);
    write_int (f, pe->size);
    fwrite (pe->data, pe->size, 1, f);
    return (TRUE);
}

/*
 * Free a preference entry
 */

void free_pref (PREFERENCE_ENTRY * pe)
{
    if (!pe)
	return;
    if (pe->name)
	free (pe->name);
    if (pe->data)
	free (pe->data);
}

PREFERENCE_ENTRY *read_pref (FILE * f)
{
    PREFERENCE_ENTRY *pe = calloc (1, sizeof (*pe));
    if (!pe)
	return (NULL);
    pe->name = read_text (f);
    if (pe->name && *pe->name)
    {
	pe->size = read_int (f);
	if (pe->size && !feof (f))
	{
	    if (!(pe->data = calloc (1, pe->size)))
		return (NULL);
	    if (fread (pe->data, pe->size, 1, f) != 1)
	    {
		printf ("Preferences item %s corrupted\n", pe->name);
		free (pe->data);
		free (pe);
		return (NULL);
	    }
	    return (pe);
	}
	else
	{
	    free (pe);
	    return (NULL);
	}
    }
    else
    {
	free (pe);
	return (NULL);
    }
}

PREFERENCES *PR_new_database (char *filename)
{
    PREFERENCES *prefs = calloc (1, sizeof (*prefs));
    strcpy (prefs->name, filename);
    prefs->valid = TRUE;
    return (prefs);
}

PREFERENCES *PR_open_database (char *filename)
{
    FILE *f;
    TY_int finished = FALSE;
    PREFERENCES *prefs = NULL;
    PREFERENCE_ENTRY *pe;

    if (!(f = fopen (filename, "r")))
	return (NULL);

    check_file_id (f, filename, PR_MAGIC_NO);

    prefs = calloc (1, sizeof (*prefs));

    while (!finished && !feof (f))
    {
	if (!(pe = read_pref (f)))
	{
	    finished = TRUE;
	}
	else
	{
	    add_preference (prefs, pe);
	}
    }
    fclose (f);
    prefs->valid = TRUE;
    strcpy (prefs->name, filename);
    return (prefs);
}

TY_int PR_save_database_as (PREFERENCES * prefs, char *filename)
{
    FILE *f;
    TY_int finished = FALSE;
    PREFERENCE_LIST *pl;

    if (!prefs || !prefs->valid)
	return (FALSE);

    if (!(f = smart_fopen (filename, "w", "Cannot open preferences database %s for write")))
    {
	return (FALSE);
    }

    put_file_id (f, PR_MAGIC_NO);

    pl = prefs->pref_list;
    while (!finished && pl)
    {
	if (!write_pref (pl->data, f))
	    finished = TRUE;
	pl = pl->next;
    }
    prefs->changed = FALSE;
    fclose (f);
    return (TRUE);
}

TY_int PR_save_database (PREFERENCES * prefs)
{
    return PR_save_database_as (prefs, prefs->name);
}

TY_int PR_close_database (PREFERENCES * prefs)
{
    PREFERENCE_LIST *pl, *npl;

    if (!prefs || !prefs->valid)
	return (FALSE);

    pl = prefs->pref_list;
    while (pl)
    {
	free_pref (pl->data);
	npl = pl->next;
	free (pl);
	pl = npl;
    }
    return (TRUE);
}

PREFERENCE_ENTRY *PR_get_item (PREFERENCES * prefs, char *name)
{
    PREFERENCE_LIST *pl;

    if (!prefs || !prefs->valid || !name || !*name)
	return (NULL);

    pl = prefs->pref_list;
    while (pl)
    {
	if (!strcmp (pl->data->name, name))
	    return (pl->data);
	pl = pl->next;
    }
    return (NULL);
}

void PR_mark_as_changed (PREFERENCES * prefs)
{
    prefs->changed = TRUE;
}

TY_int PR_add_item (PREFERENCES * prefs, char *name, void *data, TY_int size)
{
    PREFERENCE_ENTRY *pe;
    TY_int new = TRUE;

    if (!prefs || !prefs->valid)
	return (FALSE);

    if (pe = PR_get_item (prefs, name))
    {
	free_pref (pe);
	new = FALSE;
    }
    else
	pe = calloc (1, sizeof (*pe));

/*
 * prefs->changed = TRUE;
 */

    pe->name = calloc (1, strlen (name) + 1);
    strcpy (pe->name, name);

    pe->size = size;

    pe->data = calloc (1, size);
    memcpy (pe->data, data, size);
    if (new)
	add_preference (prefs, pe);
    return (TRUE);
}

/*
 * To make accessing the database simpler, these routines may
 * be used when only one global database is in use.
 */

static PREFERENCES *pref;

void PR_open (char *name)
{
    pref = PR_open_database (name);
    if (!pref)
	pref = PR_new_database (name);
}

void PR_mark_changed (void)
{
    PR_mark_as_changed (pref);
}
/*
 * Close the prefs DB. Ask to save changes if changed
 * - non zero return code means abort exit.
 */
TY_int PR_finish (void)
{
    static TY_int done = FALSE;
TY_int rc;

    if (done)
	return(0);

    done = TRUE;

    if ((pref && pref->changed) || PR_check_changes ())
    {
rc = multi_do_it ("Save changes to preferences ? ",NULL,
                         "_Yes|_No|_Cancel Exit");
if (rc)
{
if (rc==1)
PR_save_database (pref);
return (0);
}
if(!rc)
{
done = 0;
return (1);
}
    }
return (0);
}

PREFERENCES *PR_get_prefs (void)
{
    return (pref);
}


GENERIC_PREF *PR_init_load (TY_int sz, char *pref_name, void (*set_default) (void *), char *op_line)
{
    PREFERENCE_ENTRY *pe;
    GENERIC_PREF *np = calloc(sizeof(*np),1);
    void *new_ptr = calloc (sz, 1);
    void *old_ptr = old_ptr = calloc (sz, 1);

    pe = PR_get_item (PR_get_prefs (), pref_name);

    if (!pe || pe->size != sz)
    {
/*
 * If new prefs have been added, try to save the old prefs
 */
	(*set_default) (new_ptr);
	if (pe && pe->size != sz)
	{
/*
 * if (do_it ("Convert old preferences structure to new (%s)", pref_name))
 */
	    {
		if (op_line)
		{
		    strcat (op_line, "\n   ");
		    strcat (op_line, pref_name);
		}
		PR_mark_changed ();
		memcpy (new_ptr, pe->data, min(sz,pe->size));
	    }
	}
	PR_add_item (PR_get_prefs (), pref_name, new_ptr, sz);
	pe = PR_get_item (PR_get_prefs (), pref_name);
    }
    if (pe)
    {
	free (new_ptr);
	new_ptr = pe->data;
	if (old_ptr)
	    memcpy (old_ptr, new_ptr, sz);
    }
    else
    {
	(*set_default) (new_ptr);
	if (old_ptr)
	    memcpy (old_ptr, new_ptr, sz);
    }
np->set_default = set_default;
np->pe = pe;
np->original = old_ptr;
    return (np);
}
